import greenfoot.*;  
import java.util.*;
import java.awt.*;
import java.lang.*;

public class BumpMap extends Actor {
    private GreenfootImage colorImg = new SourceImage().getImage();
    private GreenfootImage heightImg = new BumpMapImage().getImage();
    
    private int[][] heightMap = new int[380][380];
    private int[][] lightMap = new int[380][380];
    
    public BumpMap() {   
        calcLightMap();
        calcHeightMap();   
    }
    
    public void addedToWorld(World _world) {
        this.setImage(colorImg);
    }
    
    private void calcLightMap() {
        for (int y = 0; y < 380; y++) {
            for (int x = 0; x < 380; x++) {
                // calculate the distance from the center of the lightmap to each pixel.
                // invert it so we have 255 in the center and 0 at the edges.
                // make sure the numbers are in the range [0-255]
                // this results in something that looks like a ball
                int d = constrain((int)(255 - 255 * dist(190, 190, x, y) / 190), 0, 255);
                lightMap[x][y] = d;
            }
        }
    }
    
    private int constrain(int _value, int _start, int _end) {
        if (_value<_start) {
            return _start;
        } else if (_value>_end) {
            return _end;
        } else {
            return _value;
        }
    }
    
    private int dist(int _x1, int _y1, int _x2, int _y2) {
        double x = Math.pow((double)_x1-(double)_x2, 2);
        double y = Math.pow((double)_y1-(double)_y2, 2);
        return (int)Math.sqrt(x+y);
    }
    
    private void calcHeightMap() {
        // extract the blue channel from the bump map - and use that as height value 
        for (int y = 0; y < 380; y++) {    
            for (int x = 0; x < 380; x++) {
                // blue is in the last 8 bits, which we get with the bitmask 0xff
                heightMap[x][y] = heightImg.getColorAt(x, y).getBlue();
            }
        }
    }
   
    public void act() {
        draw();
    }    
    
    private void draw() {
        GreenfootImage image = new GreenfootImage(380,380);
        
        Date now = new Date();
        
        // move the light in harmonic curves
        int lightX = (int)(200 * Math.cos(now.getTime()/1000.0));
        int lightY = (int)(200 * Math.sin(now.getTime()/3000.0));
        
        // we skip a 1 pixel border of the screen to accomodate for the gradient calculation
        for (int y = 1; y < 379; y++) {
            for (int x = 1; x < 379; x++) {
                // calculate the gradient at each pixel, by comparing the height of neighbouring pixels
                int dx = heightMap[x-1][y] - heightMap[(x+1)][y];
                int dy = heightMap[x][(y-1)] - heightMap[x][(y+1)];
                
                // calculate where to read the light intensity from the light map
                // start with current pixel position
                // offset by the light position
                // and offset it further based on the gradient values (dx, dy)
                // we double the gradient to make edges more visible.
                // try multiplying with four instead for more dramatic bumps :-)
                // finally make sure to clamp values to within light map range [0-379]
                int lx = constrain(lightX + x + dx * 2, 0, 379);
                int ly = constrain(lightY + y + dy * 2, 0, 379);
                
                // lookup light intensity from the light map and add a small amount of ambient light
                int intensity = lightMap[lx][ly] + 16;
                
                // read the pixel color form the color map
                Color p = colorImg.getColorAt(x, y);     

                // multiply light intensity with each color component and clamp to range [0-255]
                int r = Math.min(255, (int)(p.getRed() * intensity / 256.0));
                int g = Math.min(255, (int)(p.getGreen() * intensity / 256.0));
                int b = Math.min(255, (int)(p.getBlue() * intensity / 256.0));
                
                image.setColorAt(x, y, new Color(r, g, b));
            }
        }  
        this.setImage(image);
    }

}